{$R-}    {Range checking off}
{$B+}    {Boolean complete evaluation on}
{$S+}    {Stack checking on}
{$I+}    {I/O checking on}
{$N-}    {No numeric coprocessor}
//{$M 65500,16384,655360} {Turbo 3 default stack and heap}

Unit Mouse;

Interface

Uses
   Dos,
   Crt;

{ --------------------------------------------------------------------
  TBOMOUSE.PAS - Library of Microsoft mouse routines
  Turbo Pascal Tutor 2.0
  Copyright (c) 1986 by Borland International, INC.
  --------------------------------------------------------------------

  This file contains a library of useful routines for using the Microsoft
  Mouse from within a Turbo Pascal program.
}

const
  MouseInt = $33;

var
    MouseCol,                       { column mouse points to }
    MouseRow,                       { row mouse points to }
    ButtonStat,                     { if mouse button pressed }
    ButCount1,                      { number of presses of 1st mouse button }
                                    { since last call of mouse procedure }
    ButCount2,                      { number of presses of 2nd mouse button }
                                    { since last call of mouse procedure }
    LastXpos,                       { Last horizontal mouse position on screen}
    LastYpos,                       { Last Vertical mouse position on screen }
    MickeyCol,                      { current mouse column }
    MickeyRow    : Integer;         { current mouse row }

function MouseInstalled : boolean;

procedure HideCursor;

procedure GetMouse(var X, Y, Button : integer);

procedure SetMouse(X, Y : integer);

procedure GetButtonPress(Button : integer;
        var ButtonStatus, Count : integer;
               var LastX, LastY : Integer);

procedure SetMinMaxX(Min, Max : integer);

procedure SetMinMaxY(Min, Max : integer);

procedure SetMickeyRatio(Rx, Ry : integer);

Implementation

function MouseInstalled : boolean;
{ Reset mouse and see if mouse is installed - this function will return true
  if the mouse is installed, false otherwise.
}
var
  Reg : Registers;
begin
  with Reg do
  begin
    AX := 0;
    Intr(MouseInt, Reg);
    MouseInstalled := boolean(AX);
  end; { with }
end;

procedure HideCursor;
{ Hides the mouse cursor }
var Reg : Registers;
begin
  Reg.AX := 2;
  Intr(MouseInt, Dos.Registers(Reg));
end;

procedure GetMouse(var X, Y, Button : integer);
{ Finds the mouse position and button status }
var Reg : Registers;
begin
  with Reg do
  begin
    AX := 3;
    Intr(MouseInt, Dos.Registers(Reg));
    Button := BX;
    X := CX;
    Y := DX;
  end; { with }
end;

procedure SetMouse(X, Y : integer);
{ Moves the cursor to the specified X and Y position }
var Reg : Registers;
begin
  with Reg do
  begin
    AX := 4;
    CX := X;
    DX := Y;
    Intr(MouseInt, Dos.Registers(Reg));
  end;
end;

procedure GetButtonPress(Button : integer;
        var ButtonStatus, Count : integer;
               var LastX, LastY : Integer);
var Reg : Registers;
{ For a selected button, this procedure finds the button status, the count of
  button presses since the last call to this function, and the X and Y
  positions of the cursor at the last press of the button.
}
begin
  with Reg do
  begin
    AX := 5;
    BX := Button;
    Intr(MouseInt, Dos.Registers(Reg));
    ButtonStatus := AX;
    Count := BX;
    LastX := CX;
    LastY := DX;
  end; { with }
end;

procedure SetMinMaxX(Min, Max : integer);
{ Sets the minimum and maximum horizontal cursor positions on the screen }
var Reg : Registers;
begin
  with Reg do
  begin
    AX := 7;
    CX := Min;
    DX := Max;
    Intr(MouseInt, Dos.Registers(Reg));
  end; { with }
end;

procedure SetMinMaxY(Min, Max : integer);
{ Sets the minimum and maximum vertical cursor positions on the screen }
var Reg : Registers;
begin
  with Reg do
  begin
    AX := 8;
    CX := Min;
    DX := Max;
    Intr(MouseInt, Dos.Registers(Reg));
  end;
end;

procedure SetMickeyRatio(Rx, Ry : integer);
{ Sets the mickey to pixel ratio }
var Reg : Registers;
begin
  with Reg do
  begin
    AX := 15;
    CX := Rx;
    DX := Ry;
    Intr(MouseInt, Dos.Registers(Reg));
  end; { with }
end;

end.
